/* main.c, version 3.4.0

  Copyright (C) 2001-2007 Daniel Vollmer

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Daniel Vollmer <maven@maven.de>

*/

/** \mainpage Wavelet Image Compression Library
\author Daniel Vollmer <code@maven.de>
\date \ref changelog "2007-07-10"
\attention <a href="http://www.maven.de/code/wavelet/">Online documentation</a>,
	<a href="http://www.maven.de/code/wavelet/Kompressor.zip">Kompressor.app</a> (Mac OS 10.4 Universal),
	<a href="http://www.maven.de/code/wavelet/wavelet.tar.bz2">source code</a>,
	and	<a href="http://www.maven.de/code/repos/wavelet/">Darcs repository</a>
\section kompressor Kompressor.app
There now is a graphical user-interface for Mac OS 10.4 called <em>Kompressor.app</em>,
which can be used to easily create, inspect and modify ".wko" files. It exercises
most features of the library and can open files from the <tt>wako</tt> command-line
compressor (of the same library version). Feedback on the application and its UI
is much appreciated, it's my first proper try of a Mac OS / Cocoa application.

\section main_cmd Example code
The included command-line application (in main.c) (badly) illustrates the use of
the rest of the library. It can be used to compress and decompress .raw images
and handles color-space conversion by remembering the colorspace of each channel
in its retained data.

To build an executable on Mac OS or Linux use the following:
<tt>gcc -O2 -W -Wall -lm *.c -o wako</tt>

Many command-line parameters (such as \c -offset and \c -mse for example) can be
given repeatedly and are then set for the current channel and onwards until the
option is changed again, for example <tt>./wako -y 1 -i test_r.raw -mse 2.0 -i
test_g.raw -mse 4.0 -i test_b.raw</tt> uses a maximum MSE of 2.0 for the Y
channel (as we specified color-space conversion with \c -y), and 4.0 for the two
color channels.

\section main_own Usage in your own program
The functions provided by the library usually work on channels, e.g. the red
channel of an RGB image, or the Y channel of a YCrCb image. By default, it
supports up to 16 channels in a single file, all of which have to have matching
dimensions. The compile-time configuration and types can be changed in
config.h to allow for higher bit-depths or more channels, etc.

\subsection own_comp Compression
To compress an image, you simply do the following:
-# Create a reorder-table that is used for permuting coefficients for the size
   of channels you are about to compress with wv_create_reorder_table().
-# Initialize each channel with wv_channel_compact() to obtain a t_wv_channel
   handle. You can then free the original data.
-# Free the reorder-table with wv_free_reorder_table() if you won't
   need it again for channels of the same size.
-# Optionally, request information on whether the given settings will work and
   how much error the would approximately yield with wv_query_scheduler().
-# Write the channels to a t_bit_file with wv_encode(), giving the requested
   maximum errors and / or size as parameters.
-# Free the t_wv_channel handles with wv_channel_free().

\subsection own_decomp Decompression
To decompress an image again, you essentially reverse the above process:
-# Read a set of t_wv_channel from a t_bit_file with wv_decode(). You can either
   manually read the header (if you want to know the dimensions beforehand for
   example) and provide it to wv_decode(), or let wv_decode() read it.
-# Create a reorder-table with wv_create_reorder_table() with the dimensions
   from the t_wv_channel (unless you already have a table for that size or one
   for an exact multiple of the dimensions used).
-# Use wv_channel_uncompact() to get the original image data back. You can then
   free the t_wv_channel handle.
-# Free the reorder-table with wv_free_reorder_table() (if not needed for
   anything else).

\section further Further Reading
For more information on how the code operate, have a look at the following
pages, whose order is similar to the actual coding process.
-# \ref overview
-# \ref transform
-# \ref reorder
-# \ref coding

\section license License
This code is licensed under a zlib-license, which can be found at the top of
each source-file.
**/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <float.h> /* for DBL_MIN */
#include "io.h"
#include "utility.h"

// loads a raw gray-scale image and convert the unsigned chars to bytes
static wv_pel* raw_load(const char *fname, const int header, const int width, const int height)
{
	FILE* in = fopen(fname, "rb");
	wv_pel* out = NULL;

	if (in)
	{
		unsigned char* tmp_row;
		int i, j, xpitch, ypitch;
		int k = 0;

		xpitch = 1 << wv_log2i(width - 1);
		ypitch = 1 << wv_log2i(height - 1);

		out = malloc(xpitch * ypitch * sizeof *out);
		tmp_row = malloc(width * sizeof *tmp_row);
		fseek(in, header, SEEK_SET);
		for (i = 0; i < height; i++)
		{
			k = fread(tmp_row, sizeof *tmp_row, width, in);
			for (j = 0; j < k; j++)
				out[i * xpitch + j] = tmp_row[j]; // convert to wv_pel
		}
		if (k != width)
		{ // couldn't read all of it
			free(out);
			out = NULL;
		}
		else
		{
			for (i = 0; i < height; i++)
				for (j = width; j < xpitch; j++)
					out[i * xpitch + j] = out[i * xpitch + width - 1];
			for (i = height; i < ypitch; i++)
				memcpy(out + i * xpitch, out + (height - 1) * xpitch, xpitch * sizeof *out);
		}
		free(tmp_row);
		fclose(in);
	}
	return out;
}


// saves a raw gray-scale image from integers (takes max/min)
static void raw_save(const char *fname, const wv_pel* src, const int width, const int height, const int pitch)
{
	FILE* out = fopen(fname, "wb");

	if (out)
	{
		unsigned char *tmp_row;
		int i, j;

		tmp_row = malloc(width * sizeof *tmp_row);
		for (i = 0; i < height; i++)
		{
			for (j = 0; j < width; j++)
				tmp_row[j] = MIN(255, MAX(0, src[i * pitch + j])); // convert to unsigned char
			fwrite(tmp_row, sizeof *tmp_row, width, out);
		}
		free(tmp_row);
		fclose(out);
	}
}

void prog_func(int ccur, int mmax, void* userdata)
{
	if (mmax > 0)
	{
		static int oldpercent = -1;
		int ofs = (((int)userdata & 0xffff) * 100) / ((int)userdata >> 16);
		int percent = ofs + (ccur * 100) / (mmax * ((int)userdata >> 16));

		if (percent != oldpercent)
		{
			oldpercent = percent;
			printf("%02i%%\b\b\b", percent);
			fflush(stdout);
		}
	}
}

#define SOME_PATH_MAX 1024

typedef struct
{
	double max_mse;
	wv_pel* data;
	int input_offset;
	int approx;
	int ycocg;
	char name[SOME_PATH_MAX];
} t_channel_params;

typedef struct
{
	const char* name_long;
	const char* name_short;
	const char* help;
	void* thing;
	unsigned char what; /* i=int, f/p=float, s=string, uppercase => global, lowercase => per channel */
} t_cmd_arg;

static void print_arg(const t_cmd_arg* a)
{
	char c = tolower(a->what);

	if (c == 'i' || c == 'b')
		printf("-%-11s[-%-4s] int    (%03i) -- %s\n", a->name_long, a->name_short, *(const int *)a->thing, a->help);
	else if (c == 's' || c == 'u')
		printf("-%-11s[-%-4s] string (nil) -- %s\n", a->name_long, a->name_short, a->help);
	else
		printf("-%-11s[-%-4s] float  (0.0) -- %s\n", a->name_long, a->name_short, a->help);
}

static void print_help(int num, const t_cmd_arg a[])
{
	int i;

	puts("Generic parameters:");
	for (i = 0; i < num; i++)
		if (isupper(a[i].what))
			print_arg(&a[i]);
	puts("Individual channel parameters:");
	for (i = 0; i < num; i++)
		if (!isupper(a[i].what))
			print_arg(&a[i]);
}

#if defined(_MSC_VER)
#define strcasecmp stricmp
#endif


int main(int argc, char *argv[])
{
	double max_bpp = 0.0;
	t_reorder_table *reorder_table = NULL;
	t_wv_channel *channel[wv_MAX_CHANNELS + 1];
	int decompred = 0;
	int num_channels = 0;
	int max_bits = 0;
	int width = 512, height = 512;
	int error_code = 0;
	int i;
	char dec_input[SOME_PATH_MAX] = "";
	char output[SOME_PATH_MAX] = "";
	t_channel_params channel_params[wv_MAX_CHANNELS + 1];
	const t_cmd_arg parser[] = {
		{"width", "w", "width of input", &width, 'I'},
		{"height", "h", "height of input", &height, 'I'},
		{"maxbits", "bits", "target rate (in total bits)", &max_bits, 'B'}, /* max_bpp = 0.0 */
		{"maxbpp", "bpp", "target rate (in bits/pixel)", &max_bpp, 'R'}, /* max_bits = 0 */
		{"output", "o", "greyscale raw or wko", output, 'S'},
		{"decompress", "d", "decompress to output", dec_input, 'S'},
		{"decompred", "dr", "size reduction (halve size N times)", &decompred, 'I'},
		{"input", "i", "greyscale channel", channel_params[0].name, 'u'}, /* num_channel++ */
		{"offset", "ofs", "skip bytes of (raw) input", &channel_params[0].input_offset, 'i'},
		{"approx", "a", "enable more approximate error computation", &channel_params[0].approx, 'i'},
		{"minpsnr", "psnr", "minimum peak signal-noise ratio", &channel_params[0].max_mse, 'p'}, /* convert to mse */
		{"maxmse", "mse", "maximum mean square error", &channel_params[0].max_mse, 'f'},
		{"ycocg", "y", "enable colourspace: 1 = YCoCg, 2 = YCoCg-R", &channel_params[0].ycocg, 'i'}
	};

	printf("Wavelet Koder %s (C) 2001-2007 by Daniel Vollmer (maven@maven.de)\n\n", wv_VERSION);

	// init to lossless
	memset(channel_params, 0, sizeof channel_params);

	if (argc <= 2 || (argc % 2) == 0)
	{
		print_help(sizeof parser / sizeof parser[0], parser);
		return 1;
	}

	for (i = 1; i < argc - 1; i++)
	{
		char* arg = argv[i];
		int j;

		if (*arg == '/' || *arg == '-')
			arg++;

		for (j = 0; j < (int)(sizeof parser / sizeof parser[0]); j++)
		{
			if (strcasecmp(arg, parser[j].name_long) == 0 || strcasecmp(arg, parser[j].name_short) == 0)
			{
				double a_double;
				int an_int, k, mink, maxk;

				if (islower(parser[j].what))
				{
					mink = MAX(0, num_channels - 1);
					maxk = wv_MAX_CHANNELS;
				}
				else
					mink = maxk = 0;

				i++;
				switch (parser[j].what)
				{ /* BIRSuipf */
					case 'B':
						max_bpp = 0.0;
					case 'I':
					case 'i':
						an_int = atoi(argv[i]);
						for (k = mink; k <= maxk; k++)
							*(int *)((t_channel_params *)parser[j].thing + k) = an_int;
					break;
					case 'R':
						max_bits = 0;
					case 'f':
					case 'p':
						a_double = atof(argv[i]);
						if (parser[j].what == 'p')
							a_double = wv_psnr_to_mse(a_double);
						for (k = mink; k <= maxk; k++)
							*(double *)((t_channel_params *)parser[j].thing + k) = a_double;
					break;
					case 'S':
						/* no length check, bad!, I know... */
						strcpy((char *)parser[j].thing, argv[i]);
					break;
					case 'u':
						strcpy((char *)((t_channel_params *)parser[j].thing + num_channels), argv[i]);
						num_channels++;
				} /* switch */
				break;
			} /* if match */
		} /* for over possible args */
		if (j >= (int)(sizeof parser / sizeof parser[0]))
		{
			fprintf(stderr, "Unknown Parameter \"%s\"!\n", arg);
			print_help(sizeof parser / sizeof parser[0], parser);
			return 1;
		}
	}

	if (max_bits == 0 && max_bpp > 0.0 && width > 0 && height > 0)
		max_bits = (int)(max_bpp * width * height);

	if (*dec_input == 0)
	{ /* compressing */
		double mse[wv_MAX_CHANNELS + 1];
		double result_mse[wv_MAX_CHANNELS + 1];
		double min_mse = DBL_MAX, max_mse = 0.0;
		int result_bits[wv_MAX_CHANNELS + 1];
		int min_bits[wv_MAX_CHANNELS + 1];
		int bits;
		int width2, height2;

		if (num_channels == 0)
		{
			fputs("No input channel(s) given!\n", stderr);
			return 1;
		}

		for (i = 0; i < num_channels; i++)
		{ /* try to load each channel */
			channel_params[i].data = raw_load(channel_params[i].name, channel_params[i].input_offset, width, height);
			if (!channel_params[i].data)
			{
				fprintf(stderr, "Could not open input channel \"%s\"!\n", channel_params[i].name);
				return 1;
			}
		}

		width2 = 1 << wv_log2i(width - 1);
		height2 = 1 << wv_log2i(height - 1);

		// go through all the channels and convert their color-space (if desired & possible)
		i = 0;
		while (i < num_channels)
		{
			if (channel_params[i].ycocg != 0)
			{
				int j, ycocg = (channel_params[i].ycocg == 2) ? 2 : 1; // actual flag

				if (num_channels - i < 3)
					ycocg = 0; // not enough channels to form colorspace
				for (j = i; j < MIN(num_channels, i + 3); j++)
					channel_params[j].ycocg = ycocg; // set for other channels of this set
				// and now actually DO IT (if we've decided we can)
				if (ycocg == 2)
				{
					printf("Converting channels [%i, %i] from RGB -> YCoCg-R...\n", i + 1, i + 3);
					wv_rgb_to_ycocgr(width2 * height2, channel_params[i+0].data,
						channel_params[i+1].data, channel_params[i+2].data);
				}
				else if (ycocg == 1)
				{
					printf("Converting channels [%i, %i] from RGB -> YCoCg...\n", i + 1, i + 3);
					wv_rgb_to_ycocg(width2 * height2, channel_params[i+0].data,
						channel_params[i+1].data, channel_params[i+2].data);
				}
				i += 3; // try looking here again
			}
			else
				i++; // try the next channel
		}

		puts("Creating reorder-table...");
		reorder_table = wv_create_reorder_table(width, height);

		printf("Compacting the channels...");
		/* init the channels */
		for (i = 0; i < num_channels; i++)
		{
			channel[i] = wv_channel_compact(width, height, channel_params[i].ycocg, channel_params[i].data,
				channel_params[i].approx, reorder_table, prog_func, (void*)((num_channels << 16) + i));
			free(channel_params[i].data); // don't need the source data anymore
			channel_params[i].data = NULL;
			mse[i] = channel_params[i].max_mse;
			min_mse = MIN(min_mse, mse[i]);
			max_mse = MAX(max_mse, mse[i]);
			min_bits[i] = 0;
		}
		if (max_bits > 0)
		{ /* we have a target bit-rate, so make all MSE targets negative / relative */
			int header_size = wv_query_header(num_channels, channel);

			for (i = 0; i < num_channels; i++)
			{
				mse[i] = -DBL_MIN - ((max_mse - min_mse) - (mse[i] - min_mse));
				if (header_size > 0 && max_bits - header_size > num_channels * 384)
					min_bits[i] = (max_bits - header_size) / (num_channels * 5); /* use minbits if we have space */
			}
		}
		puts("");

		puts("Querying the scheduler...");
		bits = wv_query_scheduler(num_channels, channel, mse, max_bits, min_bits, result_mse, result_bits); /* give a bit of info beforehand */

		if (bits >= 0)
		{
			for (i = 0; i < num_channels; i++)
				printf("%i. emse: %.03f ermse: %.03f epsnr: %.03f bits: %i\n", i + 1, result_mse[i], sqrt(result_mse[i]), wv_mse_to_psnr(result_mse[i]), result_bits[i]);
			printf("-> Achieved %.02f bits / pixel for %i data bits.\n", (double)bits / (width * height), bits);
			if (*output)
			{
				t_bit_file* bf = bit_open(output, "wb", max_bits);

				if (bf)
				{
					puts("Writing output...");
					bits = wv_encode(num_channels, channel, mse, max_bits, min_bits, bf);
					if (bits <= 0)
					{
						fputs("Could not write file!\n", stderr);
						error_code = 2;
					}
					bit_close(bf, NULL);
				}
			}
		}
		else
		{
			fputs("Could not meet the constraints!\n", stderr);
			error_code = 2;
		}

		for (i = 0; i < num_channels; i++)
			wv_channel_free(channel[i]);

		wv_free_reorder_table(reorder_table); reorder_table = NULL;
	}
	else
	{ /* decompressing */
		t_bit_file* bf = bit_open(dec_input, "rb", MAX(0, max_bits));

		if (bf)
		{
			puts("Decoding the channels...");
			num_channels = wv_decode(NULL, decompred, channel, bf);
			if (num_channels > 0)
			{
				int width2; /* next bigger pow2 */
				int height2; /* next bigger pow2 */

				/* keep for later so we can free the t_wv_channels */
				width = wv_channel_get_width(channel[0]);
				height = wv_channel_get_height(channel[0]);
				printf("width: %i height: %i channels: %i\n", width, height, num_channels);
				width2 = 1 << wv_log2i(width - 1); /* next bigger pow2 */
				height2 = 1 << wv_log2i(height - 1); /* next bigger pow2 */

				if (max_bits > 0)
					printf("Used at most %.02f bits / pixel for %i bits...\n", (double)max_bits / (width * height), max_bits);

				puts("Creating reorder-table...");
				reorder_table = wv_create_reorder_table(width, height);

				puts("Uncompacting the channels...");
				for (i = 0; i < num_channels; i++)
				{
					channel_params[i].data = malloc(width2 * height2 * sizeof *channel_params[i].data);
					channel_params[i].ycocg = wv_channel_get_data_format(channel[i]) & 3;
					wv_channel_uncompact(channel[i], reorder_table, channel_params[i].data);
					wv_channel_free(channel[i]);
				}
				wv_free_reorder_table(reorder_table);

				// go through all the channels and convert their color-space
				i = 0;
				while (i < num_channels)
				{
					if (channel_params[i].ycocg != 0)
					{
						int ycocg = channel_params[i].ycocg; // actual flag

						if (num_channels - i < 3)
							ycocg = 0; // not enough channels to form colorspace
						// and now actually DO IT (if we've decided we can)
						if (ycocg == 2)
						{
							printf("Converting channels [%i, %i] from YCoCg-R -> RGB...\n", i + 1, i + 3);
							wv_ycocgr_to_rgb(width2 * height2, channel_params[i+0].data,
								channel_params[i+1].data, channel_params[i+2].data);
						}
						else if (ycocg == 1)
						{
							printf("Converting channels [%i, %i] from YCoCg -> RGB...\n", i + 1, i + 3);
							wv_ycocg_to_rgb(width2 * height2, channel_params[i+0].data,
								channel_params[i+1].data, channel_params[i+2].data);
						}
						i += 3; // try looking here again
					}
					else
						i++; // try the next channel

				}

				for (i = 0; i < num_channels; i++)
				{
					if (*output)
					{
						static const char extension[wv_MAX_CHANNELS + 1] = "RGBA0123456789UVW";
						char buf[SOME_PATH_MAX];

						if (num_channels > 1)
							sprintf(buf, "%s.%c", output, extension[i]);
						else
							strcpy(buf, output);
						raw_save(buf, channel_params[i].data, width, height, width2);
					}
					if (decompred == 0 && channel_params[i].name[0])
					{
						wv_pel* image = raw_load(channel_params[i].name, channel_params[i].input_offset, width, height);

						if (image)
						{
							double psnr, mse;

							psnr = wv_calc_psnr(image, channel_params[i].data, width, height, width2, &mse);
							printf("%i. mse: %.03f rmse: %.03f psnr: %.03f\n", i + 1, mse, sqrt(mse), psnr);
							free(image);
						}
					}
					free(channel_params[i].data);
				}
			}
			bit_close(bf, NULL);
		}
		else
		{
			fprintf(stderr, "Could not open compressed image \"%s\"...\n", dec_input);
			error_code = 3;
		}
	}
	return error_code;
}
